// apachemod/Session.cc
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott


// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "Session.hh"

#include <string>
#include <sstream>
#include <iomanip>

#include "Exception.hh"
#include "html.hh"
#include "editscript.hh"
#include "Globals.hh"

using namespace std;
using namespace pbe;


Session::Session(int r, int c, int sb, int t, string cmd, string charset_):
  rows(r),
  cols(c),
  scrollback(sb),
  time_out(t),
  command(cmd),
  charset(charset_),
  open_req(false),
  is_open(false), 
  screen(rows,cols,scrollback),
  error(false),
  be(NULL)
{}

Session::Session(const Session& other):
  id(other.id),
  rows(other.rows),
  cols(other.cols),
  scrollback(other.scrollback),
  time_out(other.time_out),
  command(other.command),
  charset(other.charset),
  open_req(other.open_req),
  is_open(other.is_open), 
  key_fifo(other.key_fifo),
  screen(other.screen),
  dirty(other.dirty),
  error_msg(other.error_msg),
  error(other.error),
  old_html_screen(other.old_html_screen),
  last_access(other.last_access),
  be(other.be)
{
  // assert(!other.screen_lock.locked);
}


// Timeout unused sessions:

void Session::touch(void)
{
  last_access = time(NULL);
}


// Check if session has a backend error to report

void Session::report_any_backend_error(void)
{
  if (error) {
    error=false;
    throw StrException(error_msg.c_str());
  }
}


void Session::close(void)
{
  open_req=false;
  Globals::singleton().notify_backend();

  for (int c=0; c<10000; c++) {
    report_any_backend_error();
    if (!is_open) {
      break;
    }
    usleep(1000);
  }

  if (is_open) {
    throw StrException("Session failed to close");
  }
}


void Session::send(std::string k)
{
  if (!is_open) {
    throw StrException("not open");
  }

  if (!k.empty()) {
    for (unsigned int i=0; i<k.length(); i++) {
      key_fifo.push(k[i]);
    }
    Globals::singleton().notify_backend();
  }

  touch();
}


std::string Session::rcv(void)
{
  if (!is_open) {
    throw StrException("Session is closed");
  }

  {
    Mutex::scoped_lock l(screen_lock);
    if (!dirty && is_open && !error) {
      struct boost::shmem::xtime timeout;
      xtime_get(&timeout, boost::shmem::TIME_UTC);
      timeout.sec += 10;
      dirty_condition.timed_wait(l,timeout);
    }
  }
  touch();

  report_any_backend_error();

  if (!is_open) {
    throw StrException("session closed while waiting");
  }

  // A short delay here will let the terminal get ahead of us if it
  // wants to.
  usleep(10000);

  ucs4_string html_screen;
  {
    Mutex::scoped_lock l(screen_lock);
    bool was_dirty = dirty;
    dirty=false;
    if (!was_dirty) {
      return "n";
    }
    html_screen = htmlify_screen(screen);
  }

  ucs4_string ucs4_editscript = make_editscript(old_html_screen.c_str(),html_screen);
  old_html_screen=html_screen.c_str();

  Iconver<valid,ucs4_char,char> ucs4_to_utf8(UCS4_NATIVE,"UTF-8");
  string utf8_editscript = ucs4_to_utf8(ucs4_editscript);

  return utf8_editscript;
}



