// apachemod/Session.hh
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005-2006 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#ifndef Session_hh
#define Session_hh

#include <queue>

#include "Screen.hh"
#include "SessionId.hh"

#include "Iconver.hh"


#ifdef USE_SHMEM

#include "SharedObject.hh"

#include <boost/shmem/sync/shared_mutex.hpp>
#include <boost/shmem/sync/shared_condition.hpp>

namespace shared_containers = SharedObject;

typedef boost::shmem::shared_mutex Mutex;
typedef boost::shmem::shared_condition Condition;

#else

#include <vector>
#include <string>
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>

namespace shared_containers = std;

typedef boost::mutex Mutex;
typedef boost::condition Condition;

#endif


class Session {
public:

  // Session objects are stored in the shared memory, and are accessed from multiple
  // processes: the backend process, and multiple Apache request handler processes.
  // All of the content of the Session object should be suitable for shared-memory,
  // i.e. POD or special shared-safe containers.
  // Per-session state that is used only by the backend process is stored in a
  // Backed object, which need not be shared-safe.  A pointer, which is valid only
  // in the backend process's address space, points from the Session object to the
  // Backend object.
  // The set of Session objects is stored in a shared-safe map<>.  This requires that
  // the Session is copy-constructable, and can be safely copied (at least, when it is
  // inserted into the map; is there any further copying after that?).

  const SessionId id;
  const int rows;
  const int cols;
  const int scrollback;
  const int time_out;
  const shared_containers::string command;
  const shared_containers::string charset;

  volatile bool open_req;
  volatile bool is_open;
  /*volatile*/ std::queue<char, shared_containers::deque<char> > key_fifo;

  typedef /*volatile*/ Screen screen_t;
  screen_t screen;
  volatile bool dirty;
  Condition dirty_condition;
  /*volatile*/ shared_containers::string error_msg;
  volatile bool error;
  shared_containers::ucs4_string old_html_screen;
  volatile time_t last_access;
  struct Backend* be;

  Session(int r, int c, int sb, int t, std::string cmd, std::string charset_);
  Session(const Session& other);
  ~Session() {};

  Mutex screen_lock;
  void touch(void);
  void report_any_backend_error(void);
  void send(std::string k);
  std::string rcv(void);
  void close(void);
};


#endif
