// apachemod/anyterm.cc
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott


// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include <unistd.h>
#include <signal.h>
#include <syslog.h>

#include <sstream>

#include "config.hh"
#include "anyterm.h"
#include "backend.h"
#include "html.hh"
#include "editscript.hh"
#include "Exception.hh"
#include "Globals.hh"
#include "HttpRequest.hh"
#include "auto_CgiParams.hh"

#include <boost/shmem/named_shared_object.hpp>


using namespace std;
using namespace pbe;


#define ENABLE_STATUS 0  /* Enable ?status URI */




void spawn_rote_process(void)
{
  pid_t pgid = getpgid(0);

  int rc = fork();
  if (rc) {
    // Parent process
    return;
  }

  // Child process

  // Put child process in same process group as parent.  This is
  // important, because Apache kills all its child processes by
  // signalling its process group.
  int pgrc = setpgid(0,pgid);
  if (pgrc==-1) {
    throw_ErrnoException("setpgid()");
  }

  try {
    try {
      backend_process();
    }
    RETHROW_MISC_EXCEPTIONS;
  }
  catch(Exception& E) {
    ostringstream s;
    E.report(s);
    syslog(LOG_ERR,"Anyterm backend process failed: %s",s.str().c_str());
    exit(1);
  }
  catch(...) {
    syslog(LOG_ERR,"Anyterm backend process failed, reason unknown");
    exit(1);
  }
}


// The shared memory region:

char* shm_name;
bool shm_opened;
static boost::shmem::named_shared_object shm_segment;


void setup_backend_shmem(void)
{
  // If this is the first time an Anyterm request has been received,
  // create the shared memory segment and spawn the backend process.
  // If this is the first time an Anyterm request has been received by
  // this process, attach to the shared memory segment.
  // If Anyterm requests have previously been received by this process,
  // do nothing.

  if (shm_opened) {
    // Not first in this process
    return;
  }


  shm_opened = shm_segment.open(shm_name);
  if (!shm_opened) {
    if (errno==ENOENT) {
      spawn_rote_process();
      do {
	usleep(1000);
	shm_opened = shm_segment.open(shm_name);
      } while (!shm_opened && errno==ENOENT);
    }
  }
  if (!shm_opened) {
    throw_ErrnoException("shm_segment.open()");
  }

  SharedObject::set_segment(shm_segment);
  Globals::set_singleton(shm_segment.find_or_construct<Globals>("globals")());
}



// Handle Apache requests:

string int_to_string(int i)
{
  char b[32];
  snprintf(b,sizeof(b),"%d",i);
  return b;
}




static string anyterm_open(string command, int rows, int cols, int scrollback, string charset)
{
  Session& ses = Globals::singleton().create_session(command, rows, cols, scrollback, ANYTERM_TIMEOUT, charset);
  return ses.id.str();
}


static void anyterm_close(SessionId id)
{
  Globals::singleton().get_session(id).close();
}


static response_t txt_resp(string s)
{
  return make_pair("text/plain; charset=\"UTF-8\"",s);
}

response_t process_request(const HttpRequest& request, string command)
{
  CgiParams params = auto_CgiParams(request);

  // Parse the arguments and call the appropriate function
  string action = params.get("a");

  try {
    try {

    if (action=="open") {
      return txt_resp(anyterm_open(command,
				   params.get_as<int>("rows",25),
				   params.get_as<int>("cols",80),
				   params.get_as<int>("sb",0),
                                   params.get("ch","ASCII")));
    }

    SessionId id(params.get("s"));

    if (action=="close") {
      anyterm_close(id);
      return txt_resp("");
      
    } else if (action=="send") {
      string k = params.get("k");
      Globals::singleton().get_session(id).send(k);
      return txt_resp("");

    } else if (action=="rcv") {
      return txt_resp(Globals::singleton().get_session(id).rcv());
      
    } else {
      throw StrException("invalid parameters '"/*+params+"'"*/);
    }

    }
    RETHROW_MISC_EXCEPTIONS;
  }
  catch (Exception& E) {
    ostringstream s;
    E.report(s);
    return txt_resp("E"+s.str());
  }
}
